import { ifSome } from "./option";

export const enum KeyEventType {
	IfCaught,
	OnlyOnce,
	AtLeastOnce,
}
type QueueItem<T> = {
	outputValue: T;
	eventType: KeyEventType;
	polled: boolean;
};

export class Keyboard<T> {
	private keysPressed: Set<T>;
	private keyQueue: Map<T, QueueItem<T>>;
	private keyMap: Map<string, QueueItem<T>>;
	constructor() {
		this.keyQueue = new Map();
		this.keyMap = new Map();
		this.keysPressed = new Set();
		document.addEventListener("keydown", (ke) => {
			this.keyDown(ke.key);
		});

		document.addEventListener("keyup", (ke) => {
			this.keyUp(ke.key);
		});
	}

	registerKeyEvent(
		event: string | Array<string>,
		outputValue: T,
		type: KeyEventType = KeyEventType.AtLeastOnce
	): void {
		const addKey = (v: string): void => {
			this.keyMap.set(v, {
				outputValue,
				eventType: type,
				polled: false,
			});
		};
		if (typeof event === "string") {
			addKey(event);
		} else {
			event.forEach((e) => {
				addKey(e);
			});
		}
	}
	clear(): void {
		this.keyQueue.clear();
	}
	reset(): void {
		this.clear();
		this.keyMap.clear();
	}
	pollSpecificKey(key: T, callback: () => void): void {
		const v = this.keyQueue.get(key);
		ifSome(v, (vv) => {
			if (!this.keysPressed.has(vv.outputValue)) {
				if (vv.eventType === KeyEventType.IfCaught || vv.polled) {
					vv.polled = false;
					this.keyQueue.delete(vv.outputValue);
					return;
				}
			}
			if (!vv.polled && vv.eventType === KeyEventType.OnlyOnce) {
				callback();
			} else if (vv.eventType !== KeyEventType.OnlyOnce) {
				callback();
			}
			vv.polled = true;
		});
	}

	poll(): Array<T> {
		const keys: Array<T> = [];

		for (const [_, event] of this.keyQueue) {
			// Check if key in queue is still being pressed
			if (!this.keysPressed.has(event.outputValue)) {
				if (event.eventType === KeyEventType.IfCaught || event.polled) {
					event.polled = false;
					this.keyQueue.delete(event.outputValue);
					continue;
				}
			}
			if (!event.polled && event.eventType === KeyEventType.OnlyOnce) {
				keys.push(event.outputValue);
			}
			if (event.eventType !== KeyEventType.OnlyOnce) {
				keys.push(event.outputValue);
			}
			event.polled = true;
		}

		return keys;
	}

	private keyDown(key: string): void {
		const mapped = this.keyMap.get(key);
		ifSome(mapped, (m) => {
			this.keysPressed.add(m.outputValue);
			if (!this.keyQueue.has(m.outputValue))
				this.keyQueue.set(m.outputValue, m);
		});
	}
	private keyUp(key: string): void {
		const mapped = this.keyMap.get(key);
		ifSome(mapped, (m) => {
			this.keysPressed.delete(m.outputValue);
		});
	}
}
